@@ -6,7 +6,7 @@
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
+    the Free Software Foundation; either version 3 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -59,14 +59,8 @@
 
 #define MEGABYTE_SECTORS (PED_MEGABYTE_SIZE / PED_SECTOR_SIZE_DEFAULT)
 
-#define BSD_DISKMAGIC 	((u_int32_t)0x82564557)
-#define BSD_BSIZE	8192
 
-struct bsdlabel {
-  unsigned int magic;
-  char unused[128];
-  unsigned int magic2;
-};
+
 
 typedef struct {
 	time_t	last_update;
@@ -81,8 +75,10 @@ static struct option	options[] = {
 	{"list",        0, NULL, 'l'},
 	{"raw-print",	0, NULL, 'r'},
 	{"size",	1, NULL, 's'},
+#ifdef GNU_EXT
 	{"linux-fdisk", 0, NULL, 'L'},
 	{"gnu-fdisk",   0, NULL, 'G'},
+#endif
 	{"interactive",	0, NULL, 'i'},
 	{"script",	0, NULL, 'p'},
 	{"sector-units",0, NULL, 'u'},
@@ -129,7 +125,8 @@ int fdisk_list_table = 0;
 int fdisk_print_raw = 0;
 
 /* Used for -s option, NULL when there was no -s option */
-const char *fdisk_part_size = NULL;
+const char *fdisk_partsize_device = NULL;
+int fdisk_partsize_part = 0;
 
 int user_cyls = 0, user_sectors = 0, user_heads = 0, user_sectsize = 0;
 
@@ -178,6 +175,10 @@ static UICalls uiquery;
    1 = Cylinders are the default unit (Default).*/
 static int cylinder_unit = 1;
 
+/* When editing a BSD label, we remember the number of logical partitions on
+   the original disk, so we display a more plausible device name */
+static int logical_offset = 0;
+
 static void _done (PedDevice* dev);
 
 /* Timer handler and other UI functions */ 
@@ -692,7 +693,7 @@ do_mkpart (PedDisk **disk)
 		_("Do you want to create the filesystem on the partition?"))) 
 	{
 		return perform_mkpartfs (*disk, &pos, part_type, fs_type,
-		                         NULL, opts);
+		                         NULL, opts | UI_WARN_COMMIT);
 	}
 	else
 	{
@@ -835,47 +836,20 @@ do_quit (PedDisk** disk)
         exit (0);
 }
 
-static int
-is_part_type_bsd (const PedPartition* part) {
-	int type = get_disk_specific_system_type(part, NULL);
-	/* We leave these for a reference */
-	if (type == 0xa5 || type == 0xa9 || type == 0xa5 ^ 0x10 || type == 0xa9 ^ 0x10)
-		return 1;
-	return 0;
-}
-static int
-is_bsd_partition (char* device, PedSector start, PedSector s_size) {
-	int fd;
-	struct bsdlabel *lbl;
-	u_char bootarea[BSD_BSIZE];
-
-	if (!device)
-		return 0;
-
-	fd = open (device, O_RDONLY);
-	if (fd < 0)
-		return 0;
-  
-	(void)lseek (fd, (off_t)start, SEEK_SET);
-  
-	/* read in the boot block area. */
-	if (read (fd, bootarea, BSD_BSIZE) != BSD_BSIZE) 
-		return 0;
-	close (fd);
-
-	lbl = (struct bsdlabel *)&(bootarea[s_size]);
-  
-	if (lbl->magic != BSD_DISKMAGIC
-	    || lbl->magic2 != BSD_DISKMAGIC) 
-		return 0;
-
-	return 1;
-}
-
 /* FIXME: This is ugly. Won't always work. And it is untested. */
 static int
 do_edit_bsd_disklabel (PedDisk** disk)
 {
+
+	if (uiquery.need_commit) {
+	        ped_exception_throw (
+				     PED_EXCEPTION_ERROR,
+				     PED_EXCEPTION_CANCEL,
+				     _("Editing a BSD label, before writing the msdos "
+			               "partition table is not supported in GNU fdisk."));
+		return 0;
+	}
+	
 	PedPartition *part;
 	int sect_size = (*disk)->dev->sector_size;
 
@@ -890,12 +864,25 @@ do_edit_bsd_disklabel (PedDisk** disk)
 		if (part->type)
 			continue;
 		if (is_part_type_bsd(part)) {
+			/* Warn if there is a filesystem on the partition */
+			if (ped_file_system_probe(&(part->geom))) {
+				if (ped_exception_throw(PED_EXCEPTION_WARNING,
+				                        PED_EXCEPTION_YES_NO,
+				     _("There is a BSD partition on the disk "
+				       "but it seems to contain a filesystem. "
+				       "This is going to destroy it. "
+			               "Are you sure you want to continue?")) 
+					== PED_EXCEPTION_NO) {
+					ped_device_end_external_access((*disk)->dev);
+					return 0;
+				}
+			}
 			/* If there is no already created BSD label created,
 			   and there is a filesystem on the partition, OR
 			   we are in gfdisk, we warn the user about it */
-			if (!is_bsd_partition((*disk)->dev->path, 
+			else if (!is_bsd_partition((*disk)->dev->path, 
 				part->geom.start * sect_size, sect_size) && 
-				(part->type || !fdisk_compatibility_mode)) {
+				(!fdisk_compatibility_mode)) {
 				if (ped_exception_throw(PED_EXCEPTION_WARNING,
 				                        PED_EXCEPTION_YES_NO,
 				     _("There is a BSD partition on the disk, "
@@ -914,8 +901,7 @@ do_edit_bsd_disklabel (PedDisk** disk)
 		PedDevice *label_dev;
 		PedDisk *label_disk;
 		char label_path[1024];
-		snprintf(label_path,sizeof(label_path),"%s%d",
-		             (*disk)->dev->path, part->num);
+		get_partition_device(label_path, sizeof(label_path), part, 0);
 		label_dev = ped_device_get(label_path);
 		if (!label_dev || ! ped_device_open(label_dev)) {
 			printf(_("There was an error opening the *BSD "
@@ -925,16 +911,21 @@ do_edit_bsd_disklabel (PedDisk** disk)
 		}
 		printf("Reading disklabel of %s at sector %d.\n",
 			(*disk)->dev->path, part->geom.start);
+		ped_exception_fetch_all ();
 		label_disk = ped_disk_new(label_dev);
 		/* FIXME: This might ruin a partition on the disk */
+		if (!label_disk) ped_exception_catch();
+		ped_exception_leave_all();
 		if (!label_disk || strcmp(label_disk->type->name, "bsd")) {
 			label_disk = ped_disk_new_fresh (label_dev,
 					ped_disk_type_get ("bsd"));
 		}
 		in_menu = 2;
+		logical_offset = count_logical_partition(*disk);
 		fdisk_interactive_menu(&label_disk, fdisk_bsd_menu_commands, 2);
 		ped_disk_destroy(label_disk);
 		ped_device_destroy(label_dev);
+		logical_offset = 0;
 		in_menu = 0;
 	} else {
 		printf(_("There is no *BSD partition on %s.\n"),
@@ -944,12 +935,14 @@ do_edit_bsd_disklabel (PedDisk** disk)
 	}
 
 	ped_device_end_external_access((*disk)->dev);
+	uiquery.need_commit = 0;
 	return 1;
 
 }
 static int
 do_print (PedDisk **disk)
 {
+	char			buf[256];
 	PedPartition*		part;
 	int			has_extended;
 	int                     sect_size;
@@ -1002,13 +995,28 @@ do_print (PedDisk **disk)
 	has_extended = ped_disk_type_check_feature ((*disk)->type,
 			       		 PED_DISK_TYPE_EXTENDED);
 
-  	unsigned int i = 0;
-	for (i = 0; i < (strlen ((*disk)->dev->path) - 5); i++)
-  		printf (" ");
-        printf ("%s %s %+10s %+11s %+11s %+4s %+7s ", _("Device"), _("Boot"),
-		_("Start"), _("End"), _("Blocks"), _("Id"), _("System"));
-	printf ("\n");
+  	unsigned int i;
 	
+	unsigned int pathlen;
+	
+	/* TODO: Make this output the same output as util-linux fdisk(?), 
+	   possibly split into seperate functions */
+	if (fdisk_compatibility_mode && !strcmp((*disk)->type->name,"bsd")) {
+		printf ("%s %+10s %+11s %+11s %+4s %+7s ", _("#   "), 
+		      _("start"), _("end"), _("blocks"), _("id"), _("system"));
+		printf ("\n");
+	}
+	else {
+		pathlen  = strlen ((*disk)->dev->path);
+		unsigned int fix = 3 - !isdigit((*disk)->dev->path[pathlen-1]);
+		pathlen += fix;
+		for (i = 0; i < (pathlen - 5 - fix); i++)
+  		printf (" ");
+		printf ("%s %s %+10s %+11s %+11s %+4s %+7s ", _("Device"),
+			 _("Boot"), _("Start"), _("End"), _("Blocks"), _("Id"),
+			 _("System"));
+	printf ("\n");
+	}
 	PedSector total_cyl = heads * sectors;
 
 	for (part = ped_disk_next_partition (*disk, NULL); part;
@@ -1044,16 +1052,22 @@ do_print (PedDisk **disk)
 		if (!ped_partition_is_active (part))
 	       		continue;
 	
-		/* If the last character is a digit, we add 'p' after the devname */
-		i = strlen((*disk)->dev->path) - 1;	
-		printf (isdigit((*disk)->dev->path[i]) ? 
-		        "%sp%d" :"%s%d ", (*disk)->dev->path, part->num);
+
+		if (fdisk_compatibility_mode && !strcmp((*disk)->type->name,"bsd")) {
+			printf("  %c: ", 'a' + part->num - 1);
+		}
+		else {
+			get_partition_device(buf, sizeof(buf), part, logical_offset);
+			printf("%2$-*1$.*1$s", pathlen, buf);
 
 		if (ped_partition_get_flag(part,PED_PARTITION_BOOT))
 			printf("  *  ");
 		else
 			printf("     ");
 		
+		}
+
+
 		printf ("%10lld %11lld %11lu ", start, end, blocks);
 
 		int type_size;
@@ -1231,10 +1245,9 @@ do_raw (PedDisk **disk)
 static void
 print_partition_size(PedDisk *disk)
 {
-	int t = atoi(fdisk_part_size);
-	PedPartition *part = ped_disk_get_partition(disk,t);
+	PedPartition *part = ped_disk_get_partition(disk,fdisk_partsize_part);
 	if (!part) {
-		printf(_("There is no partition %d on %s\n"), t, disk->dev->path);
+		printf(_("There is no partition %d on %s\n"), fdisk_partsize_part, disk->dev->path);
 		exit(1);
 	}
 	else {
@@ -2529,7 +2542,7 @@ _parse_options (int* argc_ptr, char*** argv_ptr)
 			case 'G': fdisk_compatibility_mode = 0; break;
 			case 'L': fdisk_compatibility_mode = 1; break;
 #endif
-			case 's': fdisk_part_size = optarg; break;
+			case 's': fdisk_partsize_device = optarg; break;
 			case 't': print_partition_types(); break;
 			case 'u': ped_unit_set_default(PED_UNIT_SECTOR);
 			          cylinder_unit = 0;
@@ -2563,37 +2576,10 @@ _choose_device (int* argc_ptr, char*** argv_ptr)
 	const char*	path = NULL;
 
 	/* if we want partition size */
-	if (fdisk_part_size) {
-		/* TODO: Make this more portable */
-		int t,s;
-		char* temp;
-		t = strlen(fdisk_part_size);
-		/* We count the digits at the end */
-		for (s = 0; s < t && isdigit(fdisk_part_size[t-s-1]); s++); 
-
-		if (s == 0) {
-			printf(_("You must specify a partition\n"));
-			exit(0);
-			
-		}
-		/* For names like ad0s0, cut two chars */
-		if (t > 2 && (fdisk_part_size[t-s-1] == 's' || 
-		              fdisk_part_size[t-s-1] == 'p')
-		          && isdigit(fdisk_part_size[t-s-2]))
-		{
-			temp = calloc(sizeof(char), t-s);
-			snprintf(temp, t-s, "%s", fdisk_part_size);
-		}
-		/* Otherwise, cut one char */
-		else //if (t > 1 && !isdigit(fdisk_part_size[t-2]))
-		{
-			temp = calloc(sizeof(char), t-s+1);
-			snprintf(temp, t-s+1, "%s", fdisk_part_size);
-		}
-		
-		path = temp;
-		/* We set the pointer to point to the digits */
-		fdisk_part_size += t - s;
+	if (fdisk_partsize_device) {
+		static char buf[512];
+		fdisk_partsize_part = cut_device_partnum(buf, sizeof(buf), fdisk_partsize_device);
+		path = buf;
 	}
 	else if (*argc_ptr) {
 		path = (*argv_ptr) [0];
@@ -2756,7 +2742,7 @@ fdisk (int argc, char** argv)
 	if (!disk) 
 	        return 1;
 	/* Show the size of the partition */
-	if (fdisk_part_size)
+	if (fdisk_partsize_device)
 		print_partition_size(disk);
 
 	/* List the specified disk. */
